/*
 * 	libglacier.c - Backend C library for Glacier
 *
 * 	This file is part of Glacier.
 *
 * 	Glacier is free software: you can redistribute it and/or modify it under the terms of the
 * 	GNU General Public License as published by the Free Software Foundation, either
 * 	version 3 of the License, or (at your option) any later version.
 *
 * 	Glacier is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * 	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * 	PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * 	You should have received a copy of the GNU General Public License along with Glacier. If
 * 	not, see <https://www.gnu.org/licenses/>.
 */

#include <color.h>
#include <dirent.h>
#include <errno.h>
#include <libconfig.h>
#include <locale.h>
#include <openssl/evp.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <wchar.h>

#include "config.h"

/* Buffer and size constants */
#define BUFFER_SIZE 1024
#define MAX_CHILDREN 64
#define MAX_SIZE 256
#define PATH_MAX_SIZE 512
#define MAX_RECURSION_DEPTH 100
#define DEFAULT_PERMISSIONS 0750

/* Define LG_VERBOSE if not defined elsewhere */
#ifndef LG_VERBOSE
#define LG_VERBOSE 0
#endif

typedef unsigned int uint;
typedef unsigned char uchar;

/*
 * Global Variables
 * 
 * Descriptions are given as comments after the variable declaration.
 * DEFINED IN: globals.h
 *
 */

config_t cfg; /* Context for libconfig */
config_setting_t *setting = NULL; /* Pointer for setting */

/* Configuration variables with default values */
int GLACIER_ALLOW_SERVICES = 0; /* Declaration of GLACIER_ALLOW_SERVICES as given in glacier.cfg */
char *GLACIER_ALLOWED_LICENSES = NULL; /* Declaration of GLACIER_ALLOWED_LICENSES as given in glacier.cfg */
int GLACIER_DO_INT_CHECK = 1; /* Declaration of GLACIER_DO_INT_CHECK as given in glacier.cfg */
int GLACIER_VERBOSE = 0; /* Declaration of GLACIER_VERBOSE as given in glacier.cfg */

/* Profile variables */
const char *GLACIER_REPO = NULL; /* Declaration of GLACIER_REPO as defined in profile.cfg */
const char *GLACIER_ARCH = NULL; /* Declaration of GLACIER_ARCH as defined in profile.cfg */
const char *GLACIER_TARGET = NULL; /* Declaration of GLACIER_TARGET as defined in profile.cfg */
const char *GLACIER_LOCALDB = NULL; /* Declaration of GLACIER_LOCALDB as defined in profile.cfg */
const char *GLACIER_SYSTEM_PROFILE = NULL; /* Declaration of GLACIER_SYSTEM_PROFILE as defined in profile.cfg */

/* Required runtime files */
const char *runtime_files[] = { 
	"/etc/glacier.cfg",
	"/etc/glacier/call-hooks",
	"/etc/make.conf"
	};

FILE *expected_hash; /* File pointer for expected hash */
FILE *pkg; /* File pointer for package hash */

/*
 * END GLOBAL VARIABLES
 */

/*
 * node
 *
 * DESCRIPTION: Definition of node type.
 * DEFINED IN: data.h
 *
 */

struct node {
	char *data;
	struct node *children[MAX_CHILDREN];
	int numChildren;
	struct node *left;
	struct node *right;
};

/*
 * infolog
 *
 * DESCRIPTION: Output a stylized info message.
 * PARAMETERS: char MSG[]
 * DEFINED IN: log.h
 *
 */

void
infolog(char MSG[])
{
	if (MSG == NULL) {
		return;
	}

	printf(COL_BLUE "[i]" COL_RESET " %s\n", MSG);
	return;
}

/*
 * warnlog
 *
 * DESCRIPTION: Output a stylized warning message.
 * Parameters: char MSG[]
 * DEFINED IN: log.h
 *
 */

void
warnlog(char MSG[])
{
	if (MSG == NULL) {
		return;
	}

	printf(COL_YELLOW "[!]" COL_RESET " %s\n", MSG);
	return;
}

/*
 * errlog
 * 
 * DESCRIPTION: Output a stylized error message.
 * PARAMETERS: char MSG[]
 * DEFINED IN: log.h
 *
 */

void
errlog(char MSG[])
{
	if (MSG == NULL) {
		return;
	}

	fprintf(stderr, COL_RED "[x]" COL_RESET " %s\n", MSG);
	return;
}

/*
 * successlog
 * 
 * DESCRIPTION: Output a stylized success message.
 * PARAMETERS: char MSG[]
 * DEFINED IN: log.h
 *
 */

void
successlog(char MSG[])
{
	if (MSG == NULL) {
		return;
	}

	printf(COL_GREEN "[*]" COL_RESET " %s\n", MSG);
	return;
}

/*
 * runtime_exists
 *
 * DESCRIPTION: Check if necesary runtime files exist.
 * PARAMETERS: None.
 * DEFINED IN: runtime.h
 *
 */

int
runtime_exists(void)
{
	int f;

	for (f = 0; f < sizeof(runtime_files); f++) {
		if (access(runtime_files[f], F_OK) == 0) {
			continue;
		} else {
			/* might keep this message, might not, idk */
			/* errlog("Runtime files are missing, cannot continue with operation.");
			printf(COL_RED "[x]" COL_RESET " The following files are missing:\n");
			printf(COL_RED "[x]" COL_RESET " \t%s\n", runtime_files[f]); */
			return 0;
		}
	}
	return 1;
}

/*
 * is_process_root
 *
 * DESCRIPTION: Check if process is running as root.
 * PARAMETERS: char MSG[]
 * DEFINED IN: runtime.h
 *
 */

int
is_process_root(void)
{
	if (getuid() != 0) {
		return 0; /* process is not running as root */
	}
	else {
		return 1; /* process is running as root */
	}
}

/*
 * init_config
 *
 * DESCRIPTION: Initialize libconfig.
 * PARAMETERS: None.
 * DEFINED IN: config.h
 *
 */

int
init_config(void)
{
	config_init(&cfg);

	if (! config_read_file(&cfg, runtime_files[0])) {
		fprintf(stderr, "%s:%d - %s\n", config_error_file(&cfg),
			config_error_line(&cfg), config_error_text(&cfg));
		config_destroy(&cfg);
		return(EXIT_FAILURE);
	}

	if (LG_VERBOSE == 1) {
		infolog("Initialized libconfig");
	}

	return 0;
}

/*
 * die_config
 *
 * DESCRIPTION: Kill libconfig.
 * PARAMETERS: None.
 * DEFINED IN: config.h
 *
 */

int
die_config(void)
{
	config_destroy(&cfg);

	if (LG_VERBOSE == 1) {
		infolog("Destroyed libconfig");
	}

	return(EXIT_SUCCESS);
}

/*
 * load_all_from_config
 *
 * DESCRIPTION: Loads all settings from the Glacier config file.
 * PARAMETERS: None.
 * DEFINED IN: config.h
 *
 */

int
load_all_from_config(void)
{
	/* this is probably really ugly but it works */
	if (! config_lookup_bool(&cfg, "GLACIER_DO_INT_CHECK", &GLACIER_DO_INT_CHECK)) { return 1; }
	if (! config_lookup_bool(&cfg, "GLACIER_VERBOSE",      &GLACIER_VERBOSE)) { return 1; }
	
	return 0;
}

/*
 * load_all_from_profile
 *
 * DESCRIPTION: Loads all settings from the Glacier system profile.
 * PARAMETERS: None.
 * DEFINED IN: config.h
 *
 */

int
load_all_from_profile(void)
{
	if (! config_lookup_string(&cfg, "GLACIER_REPO", 	     &GLACIER_REPO)) { return 1; }
	if (! config_lookup_string(&cfg, "GLACIER_ARCH",             &GLACIER_ARCH)) { return 1; }
	if (! config_lookup_string(&cfg, "GLACIER_TARGET", 	     &GLACIER_TARGET)) { return 1; }
	if (! config_lookup_string(&cfg, "GLACIER_SYSTEM_PROFILE",   &GLACIER_SYSTEM_PROFILE)) { return 1; }

	return 0;
}

/*
 * load_setting_from_config
 *
 * DESCRIPTION: Load a specified setting from the Glacier config file.
 * PARAMETERS: char SETTING[]
 * DEFINED IN: config.h
 *
 */

/* int
load_setting_from_config(char SETTING[])
{} */

/*
* create_node
*
* DESCRIPTION: Create a dependency tree node.
* PARAMETERS: char *data
* DEFINED IN: data.h
*
*/

struct node
*create_node(char *data)
{
	if (data == NULL) {
		if (LG_VERBOSE == 1) { errlog("NULL data passed to create_node()"); }
		return NULL;
	}

	struct node *newNode = (struct node*)malloc(sizeof(struct node));
	if (newNode == NULL) {
		if (LG_VERBOSE == 1) { errlog("Memory allocation failed in create_node()"); }
		return NULL;
	}

	newNode->data = strdup(data);
	if (newNode->data == NULL) {
		if (LG_VERBOSE == 1) { errlog("String duplication failed in create_node()"); }
		free(newNode);
		return NULL;
	}

	newNode->numChildren = 0;
	newNode->left = NULL;
	newNode->right = NULL;
	
	for (int i = 0; i < MAX_CHILDREN; i++) {
		newNode->children[i] = NULL;
	}
	
	return newNode;
}

/*
 * free_node
 *
 * DESCRIPTION: Recursively free a node and all its children.
 * PARAMETERS: struct node *root
 * DEFINED IN: data.h
 *
 */

void
free_node(struct node *root)
{
	if (root == NULL) {
		return;
	}
	
	/* Free all children recursively */
	for (int i = 0; i < root->numChildren; i++) {
		free_node(root->children[i]);
	}
	
	/* Free data string */
	if (root->data != NULL) {
		free(root->data);
	}
	
	/* Free node itself */
	free(root);
}

/*
 * add_child
 *
 * DESCRIPTION: Add a child node to a parent node.
 * PARAMETERS: struct node *parent, struct node *child
 * DEFINED IN: data.h
 *
 */

int
add_child(struct node *parent, struct node *child)
{
	if (parent == NULL || child == NULL) {
		if (LG_VERBOSE == 1) { 
			errlog("NULL pointer passed to add_child()"); 
		}
		return 1;
	}

	if (parent->numChildren < 0 || parent->numChildren >= MAX_CHILDREN) {
		if (LG_VERBOSE == 1) { 
			errlog("Invalid numChildren value in parent node");
		}
		return 2;
	}

	if (parent->numChildren < MAX_CHILDREN) {
		parent->children[parent->numChildren++] = child;
		return 0;
	} else {
		if (LG_VERBOSE == 1) { 
			errlog("Maximum number of children exceeded");
		}
		return 3;
	}
}

/*
* print_tree
*
* DESCRIPTION: Print a dependency tree.
* PARAMETERS: struct node *root, int level
* DEFINED IN: data.h
*
*/

int
print_tree(struct node *root, int level)
{
	if (root == NULL) {
		return 0;
	}
	
	if (level < 0) {
		if (LG_VERBOSE == 1) {
			errlog("Invalid level value in print_tree()");
		}
		return 1;
	}
	
	if (level > MAX_RECURSION_DEPTH) {
		/* Safety check to prevent stack overflow from recursive calls */
		if (LG_VERBOSE == 1) {
			errlog("Maximum recursion depth exceeded in print_tree()");
		}
		return 2;
	}

	for (int i = 0; i < level; i++) {
		printf("	");
	}

	if (root->data == NULL) {
		printf("(NULL)\n");
	} else {
		printf("%s\n", root->data);
	}

	if (root->numChildren < 0 || root->numChildren > MAX_CHILDREN) {
		if (LG_VERBOSE == 1) {
			errlog("Invalid numChildren value in node");
		}
		return 3;
	}

	for (int i = 0; i < root->numChildren; i++) {
		if (print_tree(root->children[i], level + 1) != 0) {
			/* Propagate errors up the call stack */
			return 4;
		}
	}
	
	return 0;
}

/*
 * mkworkspace
 *
 * DESCRIPTION: Creates a new Glacier workspace in /tmp.
 * PARAMETERS: None.
 * DEFINED IN: pkgops.h
 *
 */

int
mkworkspace(void)
{
	const char *workspace_path = "/tmp/glacier-workspace";
	
	DIR* workspace = opendir(workspace_path);
	if (workspace) {
		/* Workspace exists */
		closedir(workspace);
		/* infolog("Not creating new workspace, valid workspace already exists."); */
		return 0;
	} else if (ENOENT == errno) {
		/* infolog("Creating new Glacier workspace..."); */
		if (mkdir(workspace_path, DEFAULT_PERMISSIONS) != 0) {
			if (LG_VERBOSE == 1) {
				errlog("in mkworkspace()");
				errlog("mkdir() failed to create workspace directory");
				fprintf(stderr, "Error: %s\n", strerror(errno));
			}
			return -1;
		}
		return 1;
	} else {
		/* Some other error occurred */
		if (LG_VERBOSE == 1) {
			errlog("in mkworkspace()");
			errlog("opendir() failed to check workspace");
			fprintf(stderr, "Error: %s\n", strerror(errno));
		}
		return -1;
	}
}

/*
 * prepare_pkg
 *
 * DESCRIPTION: Copies a package archive from the localdb to the workspace, and unpacks it.
 * PARAMETERS: char PACKAGE[]
 * DEFINED IN: pkgops.h
 *
 */

int
prepare_pkg(char PACKAGE[])
{
	if (PACKAGE == NULL) {
		printf(COL_RED "[x] " COL_RESET "Package '%s' does not exist in the local database.\n", PACKAGE);
		errlog("Ensure your local database is up to date and try again.");
		errlog("This can be done by running 'glacier-update-pkgdb' as root.");
		return 1;
	} else {
		char PKG_NEW[PATH_MAX_SIZE];
		PKG_NEW[0] = '\0'; /* Initialize the string buffer */
		snprintf(PKG_NEW, sizeof(PKG_NEW), "/tmp/glacier-workspace/%s.tar", PACKAGE);
		
		FILE *pkg_old, *pkg_new;
		char buffer[BUFFER_SIZE];
		size_t bytes_read;

		pkg_old = fopen(PACKAGE, "r");
		if (pkg_old == NULL) {
			errlog("Failed to open source package file");
			return 1;
		}
		
		pkg_new = fopen(PKG_NEW, "w");
		if (pkg_new == NULL) {
			errlog("Failed to create destination package file");
			fclose(pkg_old);
			return 1;
		}

		/* Use buffered I/O for better efficiency */
		while ((bytes_read = fread(buffer, 1, sizeof(buffer), pkg_old)) > 0) {
			if (fwrite(buffer, 1, bytes_read, pkg_new) != bytes_read) {
				errlog("Failed to write to destination file");
				fclose(pkg_old);
				fclose(pkg_new);
				return 1;
			}
		}

		fclose(pkg_old);
		fclose(pkg_new);
		
		/* Fork before exec to prevent process replacement */
		pid_t pid = fork();
		if (pid < 0) {
			errlog("Failed to fork process");
			return 1;
		} else if (pid == 0) {
			/* Child process */
			char *tar_args[] = {
				"/bin/tar", /* This should be changed to /glacier/bin/tar later on */
				"-xvf",
				PKG_NEW,
				NULL,
			};

			execvp("/bin/tar", tar_args);
			/* If we get here, execvp failed */
			fprintf(stderr, COL_RED "[x] " COL_RESET "Failed to execute tar: %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		} else {
			/* Parent process */
			int status;
			waitpid(pid, &status, 0);
			
			if (WIFEXITED(status) && WEXITSTATUS(status) != 0) {
				printf(COL_RED "[x] " COL_RESET "Error while unpacking archive for package %s.\n", PACKAGE);
				return WEXITSTATUS(status);
			}
			
			if (remove(PKG_NEW) != 0) {
				warnlog("Failed to remove temporary package file");
			}
			
			char pkg_dir[PATH_MAX_SIZE];
			pkg_dir[0] = '\0'; /* Initialize the string buffer */
			snprintf(pkg_dir, sizeof(pkg_dir), "/tmp/glacier-workspace/%s", PACKAGE);
			
			if (chdir(pkg_dir) != 0) {
				errlog("Failed to change directory");
				return 1;
			}
			
			return 0;
		}
	}
	
	/* Should never reach here, but added for completeness */
	return 1;
}

/*
 * run_make_task
 *
 * DESCRIPTION: Runs a make task in current working directory 
 * PARAMETERS: char TASK[]
 * DEFINED IN: pkgops.h
 *
 */

int
run_make_task(char TASK[])
{
	if (TASK == NULL) {
		errlog("No task specified for make");
		return 1;
	}

	/* Fork before exec to prevent process replacement */
	pid_t pid = fork();
	if (pid < 0) {
		errlog("Failed to fork process");
		return 1;
	} else if (pid == 0) {
		/* Child process */
		char *build_args[] = {
			"/bin/make", /* This should be changed to /glacier/bin/make later on */
			TASK,
			NULL
		};

		execvp("/bin/make", build_args);
		/* If we get here, execvp failed */
		errlog("Failed to execute make");
		fprintf(stderr, "Error: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	} else {
		/* Parent process */
		int status;
		waitpid(pid, &status, 0);
		
		if (WIFEXITED(status)) {
			return WEXITSTATUS(status);
		} else {
			errlog("Make process terminated abnormally");
			return 1;
		}
	}

	/* Should never reach here, but added for completeness */
	return 1;
}

/*
 * print_hash
 *
 * DESCRIPTION: Prints a specified hash string
 * PARAMETERS: unsigned char *hash, unsigned int length
 * DEFINED IN: security.h
 *
 */

int
print_hash(uchar *hash, uint length)
{
	if (hash == NULL) {
		errlog("NULL hash pointer passed to print_hash()");
		return 1;
	}

	if (length == 0) {
		warnlog("Zero-length hash passed to print_hash()");
	}

	for (uint index = 0; index < length; index++) {
		printf("%02x", hash[index]);
	}

	printf("\n");

	return 0;
}

/*
 * stash_hash
 *
 * DESCRIPTION: Stores a hash inside a string
 * PARAMETERS: unsigned char *stored_hash, unsigned char *hash, unsigned int length
 * DEFINED IN: security.h
 *
 */

int
stash_hash(char *stored_hash, unsigned int stored_hash_size, const uchar *hash, uint length)
{
	if (stored_hash == NULL || hash == NULL) {
		if (LG_VERBOSE == 1) {
			errlog("in stash_hash()");
			errlog("NULL pointer provided");
		}
		return 1;
	}
	
	if (stored_hash_size < (length * 2 + 1)) {
		if (LG_VERBOSE == 1) {
			errlog("in stash_hash()");
			errlog("Output buffer too small for hash");
		}
		return 1;
	}

	/* Clear the buffer first */
	stored_hash[0] = '\0';
	
	/* Format each byte as a 2-character hex value */
	for (uint index = 0; index < length; index++) {
		char hex[3];
		snprintf(hex, sizeof(hex), "%02x", hash[index]);
		strncat(stored_hash, hex, stored_hash_size - strlen(stored_hash) - 1);
	}

	return 0;
}

/*
 * hash_file
 *
 * DESCRIPTION: Performs a hashing operation on a file and stores the result
 * PARAMETERS: const char *filename
 * DEFINED IN: security.h
 *
 */

int
hash_file(const char *filename, unsigned char *out_hash, unsigned int *out_length)
{
	FILE *data = fopen(filename, "rb");
	if (! data && LG_VERBOSE == 1) {
		errlog("in hash_file()");
		errlog("error opening file handle");
		return 1;
	}
	else if (! data) {
		return 1;
	}

	EVP_MD_CTX *context = EVP_MD_CTX_new();
	if (! context && LG_VERBOSE == 1) {
		errlog("in hash_file()");
		errlog("error creating envelope context");
		fclose(data);
		return 2;
	}
	else if (! context) {
		fclose(data);
		return 2;
	}

	if (EVP_DigestInit_ex(context, EVP_sha256(), NULL) != 1) {
		if (LG_VERBOSE == 1) {
			errlog("in hash_file()");
			errlog("error initializing digest");
		}
		EVP_MD_CTX_free(context);
		fclose(data);
		return 3;
	}

	// Read and hash file content
	unsigned char buffer[BUFFER_SIZE];
	size_t bytes_read;
	
	while ((bytes_read = fread(buffer, 1, BUFFER_SIZE, data)) > 0) {
		if (EVP_DigestUpdate(context, buffer, bytes_read) != 1) {
			if (LG_VERBOSE == 1) {
				errlog("in hash_file()");
				errlog("error updating digest");
			}
			EVP_MD_CTX_free(context);
			fclose(data);
			return 4;
		}
	}

	if (ferror(data)) {
		if (LG_VERBOSE == 1) {
			errlog("in hash_file()");
			errlog("error reading file");
		}
		EVP_MD_CTX_free(context);
		fclose(data);
		return 5;
	}

	fclose(data);

	if (EVP_DigestFinal_ex(context, out_hash, out_length) != 1) {
		if (LG_VERBOSE == 1) {
			errlog("in hash_file()");
			errlog("error finalizing digest");
		}
		EVP_MD_CTX_free(context);
		return 6;
	}
	
	EVP_MD_CTX_free(context);

	return 0;
}
